from __future__ import print_function, absolute_import, division

import unittest
import numpy as np

from numpy.fft import fftpack
import accelerate.mkl.fftpack as mklfft


class TestMklFft(unittest.TestCase):
    @classmethod
    def generate_inputs(cls, dtype):
        inputs = [cls.make_sin_wave(dtype=dtype),
                  cls.make_sin_wave_2d(dtype=dtype),
                  cls.make_random_data(dtype=dtype, ndim=3, n=13),
                  cls.make_random_data(dtype=dtype, ndim=4, n=7),
                  cls.make_random_data(dtype=dtype, ndim=5, n=3),
                  np.ones(10, dtype=dtype),
                  np.ones(10 + 1, dtype=dtype), ]
        return inputs

    @classmethod
    def make_sin_wave(cls, dtype, n=1000):
        arrpts = np.linspace(0, np.pi * 2, n)
        return np.sin(arrpts).astype(dtype)

    @classmethod
    def make_sin_wave_2d(cls, dtype, n=1000):
        arr = cls.make_sin_wave(dtype=dtype, n=n)
        return np.vstack([arr, 1 + arr])

    @classmethod
    def make_random_data(cls, dtype, n=100, ndim=10):
        arr = np.random.random((n,) * ndim).astype(dtype)
        return arr

    def test_we_have_two_apis(self):
        self.assertTrue(mklfft.fft is not fftpack.fft)
        self.assertTrue(mklfft.ifft is not fftpack.ifft)
        self.assertTrue(mklfft.rfft is not fftpack.rfft)
        self.assertTrue(mklfft.irfft is not fftpack.irfft)
        self.assertTrue(mklfft.fftn is not fftpack.fftn)
        self.assertTrue(mklfft.ifftn is not fftpack.ifftn)
        self.assertTrue(mklfft.rfftn is not fftpack.rfftn)
        self.assertTrue(mklfft.irfftn is not fftpack.irfftn)

    def _test_fft(self, dtype):
        inputs = self.generate_inputs(dtype=dtype)

        for arrsin in inputs:
            expected = fftpack.fft(arrsin)
            got = mklfft.fft(arrsin)
            self.assertEqual(expected.shape, got.shape)
            self.assertEqual(expected.dtype, got.dtype)
            np.testing.assert_almost_equal(got, expected)

    def _test_ifft(self, dtype):
        inputs = self.generate_inputs(dtype=dtype)

        for arrsin in inputs:
            got = mklfft.ifft(mklfft.fft(arrsin))
            expected = fftpack.ifft(fftpack.fft(arrsin))
            self.assertEqual(expected.shape, got.shape)
            self.assertEqual(expected.dtype, got.dtype)
            np.testing.assert_almost_equal(got, expected)

    def test_fft_float32(self):
        self._test_fft(np.float32)

    def test_fft_float64(self):
        self._test_fft(np.float64)

    def test_fft_complex64(self):
        self._test_fft(np.complex64)

    def test_fft_complex128(self):
        self._test_fft(np.complex128)

    def test_ifft_float32(self):
        self._test_ifft(np.float32)

    def test_ifft_float64(self):
        self._test_ifft(np.float64)

    def test_ifft_complex64(self):
        self._test_ifft(np.complex64)

    def test_ifft_complex128(self):
        self._test_ifft(np.complex128)

    def _test_rfft(self, dtype):
        inputs = self.generate_inputs(dtype=dtype)

        for arrsin in inputs:
            expected = fftpack.rfft(arrsin)
            got = mklfft.rfft(arrsin)
            self.assertEqual(expected.shape, got.shape)
            self.assertEqual(expected.dtype, got.dtype)
            np.testing.assert_almost_equal(got, expected)

    def _test_irfft(self, dtype):
        inputs = self.generate_inputs(dtype=dtype)

        for arrsin in inputs:
            got = mklfft.irfft(mklfft.rfft(arrsin))
            expected = fftpack.irfft(fftpack.rfft(arrsin))
            self.assertEqual(expected.shape, got.shape)
            self.assertEqual(expected.dtype, got.dtype)
            np.testing.assert_almost_equal(got, expected)

    def test_rfft_float32(self):
        self._test_rfft(np.float32)

    def test_rfft_float64(self):
        self._test_rfft(np.float64)

    def test_irfft_float32(self):
        self._test_irfft(np.float32)

    def test_irfft_float64(self):
        self._test_irfft(np.float64)

    def _test_fftn(self, dtype):
        inputs = self.generate_inputs(dtype=dtype)

        for arrsin in inputs:
            expected = fftpack.fftn(arrsin)
            got = mklfft.fftn(arrsin)
            self.assertEqual(expected.shape, got.shape)
            self.assertEqual(expected.dtype, got.dtype)
            np.testing.assert_almost_equal(got, expected)

    def _test_ifftn(self, dtype):
        inputs = self.generate_inputs(dtype=dtype)

        for arrsin in inputs:
            got = mklfft.ifftn(mklfft.fftn(arrsin))
            expected = fftpack.ifftn(fftpack.fftn(arrsin))
            self.assertEqual(expected.shape, got.shape)
            self.assertEqual(expected.dtype, got.dtype)
            np.testing.assert_almost_equal(got, expected)

    def test_fftn_float32(self):
        self._test_fftn(np.float32)

    def test_fftn_float64(self):
        self._test_fftn(np.float64)

    def test_fftn_complex64(self):
        self._test_fftn(np.complex64)

    def test_fftn_complex128(self):
        self._test_fftn(np.complex128)

    def test_ifftn_float32(self):
        self._test_ifftn(np.float32)

    def test_ifftn_float64(self):
        self._test_ifftn(np.float64)

    def test_ifftn_complex64(self):
        self._test_ifftn(np.complex64)

    def test_ifftn_complex128(self):
        self._test_ifftn(np.complex128)

    def _test_rfftn(self, dtype):
        inputs = self.generate_inputs(dtype=dtype)

        for arrsin in inputs:
            expected = fftpack.rfftn(arrsin)
            got = mklfft.rfftn(arrsin)
            self.assertEqual(expected.shape, got.shape)
            self.assertEqual(expected.dtype, got.dtype)
            np.testing.assert_almost_equal(got, expected)

    def _test_irfftn(self, dtype):
        inputs = self.generate_inputs(dtype=dtype)

        for arrsin in inputs:
            got = mklfft.irfftn(mklfft.rfftn(arrsin))
            expected = fftpack.irfftn(fftpack.rfftn(arrsin))
            self.assertEqual(expected.shape, got.shape)
            self.assertEqual(expected.dtype, got.dtype)
            np.testing.assert_almost_equal(got, expected)

    def test_rfftn_float32(self):
        self._test_rfftn(np.float32)

    def test_rfftn_float64(self):
        self._test_rfftn(np.float64)

    def test_irfftn_float32(self):
        self._test_irfftn(np.float32)

    def test_irfftn_float64(self):
        self._test_irfftn(np.float64)


def run():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestMklFft))
    runner = unittest.TextTestRunner(verbosity=1)
    return runner.run(suite)

test_cases = (TestMklFft,)

if __name__ == '__main__':
    unittest.main()
