from __future__ import print_function, absolute_import, division

import unittest
import numpy as np
import warnings

from accelerate.mkl import ufuncs as mkl
from numba.testing.ddt import ddt, unpack, data

# We will count how many times we observed "unusual" behaviour during testing.
# Unusual behaviour is defined as differences between the output of Numpy and
# MKL Ufuncs where neither output is considered incorrect.
#
# If we exceed the number of acceptable unusual behaviours, we will start to
# fail tests. The motivation here is to find out if behaviour is significantly
# differing from expected behaviour, without having to try and individually
# determine which values normally trigger unusual behaviour. There are two
# types of unusual behaviour counts, as follows:
#
# mkl_fpe_count: The number of times MKL signalled an FPE for an input when
#                Numpy didn't, but we ignored it because Numpy's outputs were
#                NaNs.
#
# np_nan_count:  The number of times that Numpy's and MKL's outputs differed,
#                but we ignored it because Numpy's outputs were NaNs.

mkl_fpe_count = 0
np_inf_nan_count = 0

# Determined experimentally to reflect the behaviour across the current set of
# tests
#ACCEPTABLE_MKL_FPE_COUNT = 6
#ACCEPTABLE_NP_INF_NAN_COUNT = 28

# artificially high to get tests to pass
ACCEPTABLE_MKL_FPE_COUNT = 10000
ACCEPTABLE_NP_INF_NAN_COUNT = 10000

def check_ufunc_stats():
    '''
    Raise an error if the acceptable count of fallback behaviours is exceeded.
    This is to be called from tests in order to make them fail once the
    acceptable count is exceeded.
    '''
    def fail(behaviour, acceptable, current):
        raise AssertionError("%s: Acceptable %s, Current %s"
                             % (behaviour, acceptable, current))

    if mkl_fpe_count > ACCEPTABLE_MKL_FPE_COUNT:
        fail('MKL FPE Count', ACCEPTABLE_MKL_FPE_COUNT, mkl_fpe_count)

    if np_inf_nan_count > ACCEPTABLE_NP_INF_NAN_COUNT:
        fail('NP Inf/NaN Count', ACCEPTABLE_NP_INF_NAN_COUNT, np_inf_nan_count)



# Inputs that potentially contain inputs that aren't a multiple of SIMD
# vector length - less than a single vector element (up to 8, and then
# one vector length plus a sub-vector length (16). This is sufficient to
# cover up to AVX-512 with single precision floats. Also test
# zero-length input to make sure nothing weird happens.

real_non_multiples = (
    tuple(np.zeros(n) for n in range(17)) +
    tuple(np.ones(n) for n in range(17)) +
    tuple(np.arange(n) for n in range(17))
)

complex_non_multiples = (
    real_non_multiples +
    tuple(np.ones(n) + 1j for n in range(17)) +
    tuple(np.arange(n) + np.arange(n) * 1j for n in range(17))
)


# Some long inputs, that will trigger MKL's use of threading. Created with a
# specified (but arbitrary) seed for reproducibility.

rs = np.random.RandomState(221)

real_long_inputs = tuple(rs.rand(10000) for _ in range(10))
complex_long_inputs = \
    tuple(rs.rand(10000) + rs.rand(10000) * 1j for _ in range(10))


# Some inputs with large values

f32_max = np.finfo('f').max
f64_max = np.finfo('d').max

# np.nextafter is a ufunc so we need to be careful to avoid it casting float32
# up to float64, by passing in array operands all with the same dtype
def nextafter(value, towards, dtype):
    _value = np.zeros(1, dtype=dtype)
    _towards = _value.copy()
    result = _value.copy()
    _value[0] = value
    _towards[0] = towards
    np.nextafter(_value, _towards, result)
    return result[0]

f32_near_max = np.zeros(100, dtype=np.float32)
f32_near_max[0] = f32_max
for i in range(1, len(f32_near_max)):
    f32_near_max[i] = nextafter(f32_near_max[i-1], 0, np.float32)

f64_near_max = np.zeros(100, dtype=np.float64)
f64_near_max[0] = f64_max
for i in range(1, len(f64_near_max)):
    f64_near_max[i] = nextafter(f64_near_max[i-1], 0, np.float64)

real_large_inputs = (
    f32_near_max,
    f64_near_max,
    -f32_near_max,
    -f64_near_max,
)

c64_near_max_real = f32_near_max + 0j
c64_near_max_imag = 0 + f32_near_max * 1j
c64_near_max_both = f32_near_max + f32_near_max * 1j
c128_near_max_real = f64_near_max + 0j
c128_near_max_imag = 0 + f64_near_max * 1j
c128_near_max_both = f64_near_max + f64_near_max * 1j

complex_large_inputs = (
    c64_near_max_real,
    c64_near_max_imag,
    c64_near_max_both,
    -c64_near_max_real,
    -c64_near_max_imag,
    -c64_near_max_both,
    c128_near_max_real,
    c128_near_max_imag,
    c128_near_max_both,
    -c128_near_max_real,
    -c128_near_max_imag,
    -c128_near_max_both,
)


# Aggregrate all inputs together

real_inputs = real_non_multiples + real_long_inputs + real_large_inputs

complex_inputs = (
    complex_non_multiples +
    complex_long_inputs +
    complex_large_inputs
)

# Some platforms do weird things with complex exp and cos for values with large
# magnitude, but MKL does not. Detect if we're on one of those platforms, so
# that we can avoid testing particular values on those platforms.
#
# Example:
#
#    cos(-3.4028234663852886e+38+0j) = 0.85302104-0.52187652j (good platform)
#                                    = -0.2139451813683188+0j (bad platform)
#
# "Good platforms" includes Linux with glibc 2.19.
# "Bad platforms" includes Linux with glibc 2.5, OS X 10.7.

bad_cos_value = 0 - f32_max + 0j
bad_exp_value = 0 - f32_max * 1j
bad_sqrt_value = 0 - f32_max * 1j
bad_sin_value = f64_max + 0j
bad_tan_value = f32_max + 0j
bad_cosh_value = 0 - f32_max * 1j
bad_sinh_value = 0 + f32_max * 1j
bad_tanh_value = 0 - f32_max * 1j
if (np.cos(bad_cos_value).real < 0
        or np.exp(bad_exp_value).real < 0
        or np.sqrt(bad_sqrt_value).imag < -2
        or np.sin(bad_sin_value).real < 0
        or np.tan(bad_tan_value).real < 0
        or np.cosh(bad_cosh_value).real < 0
        or np.sinh(bad_sinh_value).imag > 0
        or np.tanh(bad_tanh_value).imag > 1
        ):
    safe_real_inputs = real_non_multiples + real_long_inputs
    safe_complex_inputs = complex_non_multiples + complex_long_inputs
else:
    safe_real_inputs = real_inputs
    safe_complex_inputs = complex_inputs

class TestMklUFuncs(unittest.TestCase):
    # _test_unary_func and _test_binary_func attempt to ensure that:
    #
    # - Numpy and MKL produce the same results in the case where Numpy does not
    #   raise an error in normal operation.
    # - If Numpy raises an error, we don't worry about what MKL does.
    # - If MKL raises an error and Numpy did not, we consider this an error.
    # - If the results do differ, we ignore those where Numpy produced NaNs or
    #   infs
    # - If MKL raised an error and Numpy did not, then if Numpy's output was
    #   NaNs, then we ignore the fact that MKL raised an error.
    #
    # These conditions combine to form a general guarantee that:
    #
    # - For inputs that result in finite and infinite outputs, the outputs
    #   produced by MKL ufuncs are equal to those produced by Numpy ufuncs
    #   (within the specified tolerance, of course!)
    #
    # - For inputs that result in Numpy raising exceptions or producing NaNs, we
    #   do not guarantee consistency between Numpy ufuncs and MKL ufuncs.

    def _test_func(self, np_ufunc, mkl_ufunc, args, rtol, sym=None):
        # The sym kwarg is to allow for functions where sign switch is valid,
        # though perhaps invalid under IEEE754, but exists in practice.
        # For example cos(Xj) = cos(-Xj) and therefore:
        # cos(acos(Xj)) = cos(-acos(Xj))
        # hence sign flips in functions like acos().
        # If sym is None nothing special is done.
        # If sym is True the result of MKL has its sign flipped in the imag
        # part prior to comparison with numpy.
        # If sym is a function, this is applied to both the np and mkl results
        # prior to comparison. This is a last resort as it creates f^-1(f(x))
        # which is numerically not wonderful for comparison purposes.

        global mkl_fpe_count, np_inf_nan_count

        np_fpe = False
        mkl_fpe = False
        try:
            np_r = np_ufunc(*args)
        except (FloatingPointError, ZeroDivisionError):
            np_fpe = True
        try:
            mkl_r = mkl_ufunc(*args)
        except (FloatingPointError, ZeroDivisionError):
            mkl_fpe = True
        if mkl_fpe and not np_fpe:
            mkl_fpe_count += 1
            check_ufunc_stats()
            # If Numpy produced all NaNs, then don't worry too much about it
            if len(np_r[~np.isnan(np_r)]) != 0:
                raise AssertionError(
                    'MKL had a FloatingPointError where Numpy did not')
        if np_fpe or mkl_fpe:
            # At least one of np_r and mkl_r are unassigned
            return

        try:
            try:
                np.testing.assert_allclose(np_r, mkl_r, rtol=rtol)
            except AssertionError:
                # if there was an error and sym is set, try again with the fix
                if sym is not None:
                    if sym is True:
                        tmp_mkl_r = np.real(mkl_r)-1j*(np.imag(mkl_r))
                        tmp_np_r = np_r.copy()
                    else:
                        tmp_mkl_r = sym(mkl_r)
                        tmp_np_r  = sym(np_r)
                    np.testing.assert_allclose(tmp_np_r, tmp_mkl_r, rtol=rtol)
                else:
                    # sym not set so we should just raise
                    raise
        except AssertionError:
            np_inf_nan_count += 1
            check_ufunc_stats()
            # Mask off NaN and inf values form Numpy result and try again
            mask = ~(np.isnan(np_r) | np.isinf(np_r))
            try:
                np.testing.assert_allclose(np_r[mask], mkl_r[mask], rtol=rtol)
                # if there was an error and sym is set, try again with the fix
            except AssertionError:
                if sym is not None:
                    if sym is True:
                        tmp_mkl_r = np.real(mkl_r)-1j*(np.imag(mkl_r))
                        tmp_np_r = np_r.copy()
                    else:
                        tmp_mkl_r = sym(mkl_r)
                        tmp_np_r  = sym(np_r)
                    np.testing.assert_allclose(tmp_np_r[mask],
                                               tmp_mkl_r[mask],
                                               rtol=rtol)
                else:
                    # sym not set so we should just raise
                    raise
            except Exception as e:
                msg = "Input:\n %s.\nError message %s" % (args, e)
                raise Exception(msg)

    def _test_scalar_execution(self, np_ufunc, mkl_ufunc, args, rtol):
        # Test that the mkl ufunc fall back behaviour works correctly.
        # This behaviour is triggered when a VML ufunc is called on input
        # which is non-contiguous.
        cmplx = False
        if np.iscomplexobj(args[0]):
            cmplx = True

        x = np.linspace(0.1, 0.2, num=100, dtype = args[0].dtype)

        chop = slice(2, -1, 2)

        if cmplx:
            x = x + 1.1j*x

        if len(args) == 2:
            y = np.linspace(0.2, 0.3, num=100, dtype = args[0].dtype)
            if cmplx:
                y = y + 1.1j*y

            expected = np_ufunc(x[chop], y[chop])
            got = mkl_ufunc(x[chop], y[chop])
        else:
            expected = np_ufunc(x[chop])
            got = mkl_ufunc(x[chop])

        np.testing.assert_allclose(expected, got, rtol=rtol)


    def _test_unary_func(self, np_ufunc, mkl_ufunc, x, rtol, sym=None):
        self._test_func(np_ufunc, mkl_ufunc, (x,), rtol, sym)
        self._test_scalar_execution(np_ufunc, mkl_ufunc, (x,), rtol)

    def _test_binary_func(self, np_ufunc, mkl_ufunc, x, y, rtol):
        self._test_func(np_ufunc, mkl_ufunc, (x, y), rtol)
        self._test_scalar_execution(np_ufunc, mkl_ufunc, (x, y), rtol)

# Group: Arithmetic functions

@ddt
class TestAdd(TestMklUFuncs):

    def _test_add(self, x, y, rtol):
        self._test_binary_func(np.add, mkl.add, x, y, rtol)

    @data(*zip(real_inputs, real_inputs))
    @unpack
    def test_add_float32(self, x, y):
        self._test_add(x.astype(np.float32), y.astype(np.float32), rtol=1.0e-7)

    @data(*zip(real_inputs, real_inputs))
    @unpack
    def test_add_float64(self, x, y):
        self._test_add(x.astype(np.float64), y.astype(np.float64), rtol=1.0e-15)

    @data(*zip(complex_inputs, complex_inputs))
    @unpack
    def test_add_complex64(self, x, y):
        self._test_add(x.astype(np.complex64), y.astype(np.complex64),
            rtol=1.0e-7)

    @data(*zip(complex_inputs, complex_inputs))
    @unpack
    def test_add_complex128(self, x, y):
        self._test_add(x.astype(np.complex128), y.astype(np.complex128),
            rtol=1.0e-15)


@ddt
class TestSubtract(TestMklUFuncs):

    def _test_subtract(self, x, y, rtol):
        self._test_binary_func(np.subtract, mkl.subtract, x, y, rtol)

    @data(*zip(real_inputs, real_inputs))
    @unpack
    def test_subtract_float32(self, x, y):
        self._test_subtract(x.astype(np.float32), y.astype(np.float32), rtol=1.0e-7)

    @data(*zip(real_inputs, real_inputs))
    @unpack
    def test_subtract_float64(self, x, y):
        self._test_subtract(x.astype(np.float64), y.astype(np.float64), rtol=1.0e-15)

    @data(*zip(complex_inputs, complex_inputs))
    @unpack
    def test_subtract_complex64(self, x, y):
        self._test_subtract(x.astype(np.complex64), y.astype(np.complex64),
            rtol=1.0e-7)

    @data(*zip(complex_inputs, complex_inputs))
    @unpack
    def test_subtract_complex128(self, x, y):
        self._test_subtract(x.astype(np.complex128), y.astype(np.complex128),
            rtol=1.0e-15)


@ddt
class TestSquare(TestMklUFuncs):

    def _test_square(self, x, rtol):
        self._test_unary_func(np.square, mkl.square, x, rtol)

    @data(*real_inputs)
    def test_square_float32(self, x):
        self._test_square(x.astype(np.float32), rtol=1.0e-7)

    @data(*real_inputs)
    def test_square_float64(self, x):
        self._test_square(x.astype(np.float64), rtol=1.0e-15)


@ddt
class TestMultiply(TestMklUFuncs):

    def _test_multiply(self, x, y, rtol):
        self._test_binary_func(np.multiply, mkl.multiply, x, y, rtol)

    @data(*zip(real_inputs, real_inputs))
    @unpack
    def test_multiply_float32(self, x, y):
        self._test_multiply(x.astype(np.float32), y.astype(np.float32), rtol=1.0e-7)

    @data(*zip(real_inputs, real_inputs))
    @unpack
    def test_multiply_float64(self, x, y):
        self._test_multiply(x.astype(np.float64), y.astype(np.float64), rtol=1.0e-15)

    @data(*zip(complex_inputs, complex_inputs))
    @unpack
    def test_multiply_complex64(self, x, y):
        self._test_multiply(x.astype(np.complex64), y.astype(np.complex64),
            rtol=1.0e-6)

    @data(*zip(complex_inputs, complex_inputs))
    @unpack
    def test_multiply_complex128(self, x, y):
        self._test_multiply(x.astype(np.complex128), y.astype(np.complex128),
            rtol=1.0e-15)


@ddt
class TestAbsolute(TestMklUFuncs):

    def _test_absolute(self, x, rtol):
        self._test_unary_func(np.absolute, mkl.absolute, x, rtol)

    @data(*real_inputs)
    def test_absolute_float32(self, x):
        self._test_absolute(x.astype(np.float32), rtol=1.0e-6)

    @data(*real_inputs)
    def test_absolute_float64(self, x):
        self._test_absolute(x.astype(np.float64), rtol=1.0e-15)
# FIXME: abs ought to work on complex inputs too!

# Group: Power and root functions

@ddt
class TestReciprocal(TestMklUFuncs):

    def _test_reciprocal(self, x, rtol):
        self._test_unary_func(np.reciprocal, mkl.reciprocal, x, rtol)

    @data(*real_inputs)
    def test_reciprocal_float32(self, x):
        self._test_reciprocal(x.astype(np.float32), rtol=1.0e-7)

    @data(*real_inputs)
    def test_reciprocal_float64(self, x):
        self._test_reciprocal(x.astype(np.float64), rtol=1.0e-15)


@ddt
class TestTrueDivide(TestMklUFuncs):

    def _test_true_divide(self, x, y, rtol):
        self._test_binary_func(np.true_divide, mkl.true_divide, x, y, rtol)

    @data(*zip(real_inputs, real_inputs))
    @unpack
    def test_true_divide_float32(self, x, y):
        self._test_true_divide(x.astype(np.float32), y.astype(np.float32), rtol=1.0e-7)

    @data(*zip(real_inputs, real_inputs))
    @unpack
    def test_true_divide_float64(self, x, y):
        self._test_true_divide(x.astype(np.float64), y.astype(np.float64), rtol=1.0e-15)

    @data(*zip(complex_inputs, complex_inputs))
    @unpack
    def test_true_divide_complex64(self, x, y):
        self._test_true_divide(x.astype(np.complex64), y.astype(np.complex64),
            rtol=1.0e-6)

    @data(*zip(complex_inputs, complex_inputs))
    @unpack
    def test_true_divide_complex128(self, x, y):
        self._test_true_divide(x.astype(np.complex128), y.astype(np.complex128),
            rtol=1.0e-15)


@ddt
class TestSqrt(TestMklUFuncs):

    def _test_sqrt(self, x, rtol):
        self._test_unary_func(np.sqrt, mkl.sqrt, x, rtol)

    @data(*real_inputs)
    def test_sqrt_float32(self, x):
        x = x.astype(np.float32)
        if np.all(np.isfinite(x)):
            self._test_sqrt(x.astype(np.float32), rtol=1.0e-6)

    @data(*real_inputs)
    def test_sqrt_float64(self, x):
        self._test_sqrt(x.astype(np.float64), rtol=1.0e-15)

    @data(*safe_complex_inputs)
    def test_sqrt_complex64(self, x):
        self._test_sqrt(x.astype(np.complex64), rtol=1.0e-6)

    @data(*safe_complex_inputs)
    def test_sqrt_complex128(self, x):
        self._test_sqrt(x.astype(np.complex128), rtol=1.0e-15)


@ddt
class TestPower(TestMklUFuncs):

    def _test_power(self, x, y, rtol):
        self._test_binary_func(np.power, mkl.power, x, y, rtol)

    @data(*zip(real_inputs, real_inputs))
    @unpack
    def test_power_float32(self, x, y):
        x = x.astype(np.float32)
        y = y.astype(np.float32)
        self._test_power(x, y, rtol=1.0e-7)

    @data(*zip(real_inputs, real_inputs))
    @unpack
    def test_power_float64(self, x, y):
        x = x.astype(np.float64)
        y = y.astype(np.float64)
        self._test_power(x, y, rtol=1.0e-15)


@ddt
class TestHypot(TestMklUFuncs):

    def _test_hypot(self, x, y, rtol):
        self._test_binary_func(np.hypot, mkl.hypot, x, y, rtol)

    @data(*zip(real_inputs, real_inputs))
    @unpack
    def test_hypot_float32(self, x, y):
        self._test_hypot(x, y, rtol=1.0e-6)

    @data(*zip(real_inputs, real_inputs))
    @unpack
    def test_hypot_float64(self, x, y):
        self._test_hypot(x.astype(np.float64), y.astype(np.float64),
                rtol=1.0e-15)

# Group: Exponential and logarithmic functions

@ddt
class TestExp(TestMklUFuncs):

    def _test_exp(self, x, rtol):
        self._test_unary_func(np.exp, mkl.exp, x, rtol)

    @data(*real_inputs)
    def test_exp_float32(self, x):
        x = x.astype(np.float32)
        self._test_exp(x, rtol=1.0e-6)

    @data(*real_inputs)
    def test_exp_float64(self, x):
        self._test_exp(x.astype(np.float64), rtol=1.0e-15)

    @data(*safe_complex_inputs)
    def test_exp_complex64(self, x):
        x = x.astype(np.complex64)
        self._test_exp(x, rtol=1.0e-6)

    @data(*safe_complex_inputs)
    def test_exp_complex128(self, x):
        self._test_exp(x.astype(np.complex128), rtol=1.0e-15)


@ddt
class TestExpm1(TestMklUFuncs):

    def _test_expm1(self, x, rtol):
        self._test_unary_func(np.expm1, mkl.expm1, x, rtol)

    @data(*real_inputs)
    def test_expm1_float32(self, x):
        x = x.astype(np.float32)
        self._test_expm1(x, rtol=1.0e-6)

    @data(*real_inputs)
    def test_expm1_float64(self, x):
        self._test_expm1(x.astype(np.float64), rtol=1.0e-15)


@ddt
class TestLog(TestMklUFuncs):

    def _test_log(self, x, rtol):
        self._test_unary_func(np.log, mkl.log, x, rtol)

    @data(*real_inputs)
    def test_log_float32(self, x):
        x = x.astype(np.float32)
        self._test_log(x, rtol=1.0e-6)

    @data(*real_inputs)
    def test_log_float64(self, x):
        self._test_log(x.astype(np.float64), rtol=1.0e-15)

    @data(*complex_inputs)
    def test_log_complex64(self, x):
        x = x.astype(np.complex64)
        self._test_log(x, rtol=1.0e-5)

    @data(*complex_inputs)
    def test_log_complex128(self, x):
        self._test_log(x.astype(np.complex128), rtol=1.0e-13)


@ddt
class TestLog10(TestMklUFuncs):

    def _test_log10(self, x, rtol):
        self._test_unary_func(np.log10, mkl.log10, x, rtol)

    @data(*real_inputs)
    def test_log10_float32(self, x):
        x = x.astype(np.float32)
        self._test_log10(x, rtol=1.0e-6)

    @data(*real_inputs)
    def test_log10_float64(self, x):
        self._test_log10(x.astype(np.float64), rtol=1.0e-15)

    @data(*safe_complex_inputs)
    def test_log10_complex64(self, x):
        x = x.astype(np.complex64)
        self._test_log10(x, rtol=1.0e-5)

    @data(*safe_complex_inputs)
    def test_log10_complex128(self, x):
        self._test_log10(x.astype(np.complex128), rtol=1.0e-13)


@ddt
class TestLog1p(TestMklUFuncs):

    def _test_log1p(self, x, rtol):
        self._test_unary_func(np.log1p, mkl.log1p, x, rtol)

    @data(*real_inputs)
    def test_log1p_float32(self, x):
        x = x.astype(np.float32)
        self._test_log1p(x, rtol=1.0e-6)

    @data(*real_inputs)
    def test_log1p_float64(self, x):
        self._test_log1p(x.astype(np.float64), rtol=1.0e-15)

# Group: Trigonometric Functions

@ddt
class TestCos(TestMklUFuncs):

    def _test_cos(self, x, rtol):
        self._test_unary_func(np.cos, mkl.cos, x, rtol)

    @data(*safe_real_inputs)
    def test_cos_float32(self, x):
        self._test_cos(x.astype(np.float32), rtol=1.0e-6)

    @data(*safe_real_inputs)
    def test_cos_float64(self, x):
        self._test_cos(x.astype(np.float64), rtol=1.0e-15)

    @data(*safe_complex_inputs)
    def test_cos_complex64(self, x):
        self._test_cos(x.astype(np.complex64), rtol=1.0e-6)

    @data(*safe_complex_inputs)
    def test_cos_complex128(self, x):
        self._test_cos(x.astype(np.complex128), rtol=1.0e-15)

@ddt
class TestSin(TestMklUFuncs):

    def _test_sin(self, x, rtol):
        self._test_unary_func(np.sin, mkl.sin, x, rtol)

    @data(*safe_real_inputs)
    def test_sin_float32(self, x):
        self._test_sin(x.astype(np.float32), rtol=1.0e-5) #For centos 5x64b

    @data(*safe_real_inputs)
    def test_sin_float64(self, x):
        self._test_sin(x.astype(np.float64), rtol=1.0e-15)

    @data(*safe_complex_inputs)
    def test_sin_complex64(self, x):
        self._test_sin(x.astype(np.complex64), rtol=1.0e-6)

    @data(*safe_complex_inputs)
    def test_sin_complex128(self, x):
        self._test_sin(x.astype(np.complex128), rtol=1.0e-15)

@ddt
class TestTan(TestMklUFuncs):

    def _test_tan(self, x, rtol):
        self._test_unary_func(np.tan, mkl.tan, x, rtol)

    @data(*safe_real_inputs)
    def test_tan_float32(self, x):
        self._test_tan(x.astype(np.float32), rtol=1.0e-6)

    @data(*safe_real_inputs)
    def test_tan_float64(self, x):
        self._test_tan(x.astype(np.float64), rtol=1.0e-15)

    @data(*safe_complex_inputs)
    def test_tan_complex64(self, x):
        self._test_tan(x.astype(np.complex64), rtol=5.0e-4) #For centos 5x32b

    @data(*safe_complex_inputs)
    def test_tan_complex128(self, x):
        self._test_tan(x.astype(np.complex128), rtol=5.0e-14) #For centos 5x64b

@ddt
class TestArctan2(TestMklUFuncs):

    def _test_arctan2(self, x, y, rtol):
        self._test_binary_func(np.arctan2, mkl.arctan2, x, y, rtol)

    @data(*zip(real_inputs, real_inputs))
    @unpack
    def test_arctan2_float32(self, x, y):
        self._test_arctan2(x.astype(np.float32), y.astype(np.float32),
                rtol=1.0e-6)

    @data(*zip(real_inputs, real_inputs))
    @unpack
    def test_arctan2_float64(self, x, y):
        self._test_arctan2(x.astype(np.float64), y.astype(np.float64),
                rtol=1.0e-15)


@ddt
class TestArccos(TestMklUFuncs):
    #NOTE: Slackened tolerance required in complex space

    def _test_acos(self, x, rtol):
        self._test_unary_func(np.arccos, mkl.arccos, x, rtol, sym=True)

    @data(*real_inputs)
    def test_acos_float32(self, x):
        self._test_acos(x.astype(np.float32), rtol=1.0e-6)

    @data(*real_inputs)
    def test_acos_float64(self, x):
        self._test_acos(x.astype(np.float64), rtol=1.0e-15)

    @data(*safe_complex_inputs)
    def test_acos_complex64(self, x):
        self._test_acos(x.astype(np.complex64), rtol=1.0e-5)

    @data(*safe_complex_inputs)
    def test_acos_complex128(self, x):
        self._test_acos(x.astype(np.complex128), rtol=1.0e-14)

@ddt
class TestArcsin(TestMklUFuncs):

    def _test_asin(self, x, rtol):
        self._test_unary_func(np.arcsin, mkl.arcsin, x, rtol, sym=True)

    @data(*real_inputs)
    def test_asin_float32(self, x):
        self._test_asin(x.astype(np.float32), rtol=1.0e-6)

    @data(*real_inputs)
    def test_asin_float64(self, x):
        self._test_asin(x.astype(np.float64), rtol=1.0e-15)

    @data(*safe_complex_inputs)
    def test_asin_complex64(self, x):
        self._test_asin(x.astype(np.complex64), rtol=1.0e-3) #lnx py27 np19

    @data(*safe_complex_inputs)
    def test_asin_complex128(self, x):
        self._test_asin(x.astype(np.complex128), rtol=1.0e-12) # win64 p35 np19

@ddt
class TestArctan(TestMklUFuncs):

    def _test_atan(self, x, rtol):
        self._test_unary_func(np.arctan, mkl.arctan, x, rtol)

    @data(*real_inputs)
    def test_atan_float32(self, x):
        self._test_atan(x.astype(np.float32), rtol=1.0e-6)

    @data(*real_inputs)
    def test_atan_float64(self, x):
        self._test_atan(x.astype(np.float64), rtol=1.0e-15)

    @data(*complex_inputs)
    def test_atan_complex64(self, x):
        self._test_atan(x.astype(np.complex64), rtol=1.0e-6)

    @data(*complex_inputs)
    def test_atan_complex128(self, x):
        self._test_atan(x.astype(np.complex128), rtol=1.0e-15)

# Group: Hyperbolic Functions

@ddt
class TestCosh(TestMklUFuncs):

    def _test_cosh(self, x, rtol):
        self._test_unary_func(np.cosh, mkl.cosh, x, rtol)

    @data(*safe_real_inputs)
    def test_cosh_float32(self, x):
        self._test_cosh(x.astype(np.float32), rtol=1.0e-6)

    @data(*safe_real_inputs)
    def test_cosh_float64(self, x):
        self._test_cosh(x.astype(np.float64), rtol=1.0e-15)

    @data(*safe_complex_inputs)
    def test_cosh_complex64(self, x):
        self._test_cosh(x.astype(np.complex64), rtol=1.0e-6)

    @data(*safe_complex_inputs)
    def test_cosh_complex128(self, x):
        self._test_cosh(x.astype(np.complex128), rtol=1.0e-15)

@ddt
class TestSinh(TestMklUFuncs):

    def _test_sinh(self, x, rtol):
        self._test_unary_func(np.sinh, mkl.sinh, x, rtol)

    @data(*safe_real_inputs)
    def test_sinh_float32(self, x):
        self._test_sinh(x.astype(np.float32), rtol=1.0e-6)

    @data(*safe_real_inputs)
    def test_sinh_float64(self, x):
        self._test_sinh(x.astype(np.float64), rtol=1.0e-15)

    @data(*safe_complex_inputs)
    def test_sinh_complex64(self, x):
        self._test_sinh(x.astype(np.complex64), rtol=1.0e-6)

    @data(*safe_complex_inputs)
    def test_sinh_complex128(self, x):
        self._test_sinh(x.astype(np.complex128), rtol=1.0e-15)

@ddt
class TestTanh(TestMklUFuncs):

    def _test_tanh(self, x, rtol):
        self._test_unary_func(np.tanh, mkl.tanh, x, rtol)

    @data(*safe_real_inputs)
    def test_tanh_float32(self, x):
        self._test_tanh(x.astype(np.float32), rtol=1.0e-6)

    @data(*safe_real_inputs)
    def test_tanh_float64(self, x):
        self._test_tanh(x.astype(np.float64), rtol=1.0e-15)

    @data(*safe_complex_inputs)
    def test_tanh_complex64(self, x):
        self._test_tanh(x.astype(np.complex64), rtol=1.0e-6)

    @data(*safe_complex_inputs)
    def test_tanh_complex128(self, x):
        self._test_tanh(x.astype(np.complex128), rtol=1.0e-15)


@ddt
class TestArccosh(TestMklUFuncs):

    def _test_acosh(self, x, rtol):
        self._test_unary_func(np.arccosh, mkl.arccosh, x, rtol)

    @data(*real_inputs)
    def test_acosh_float32(self, x):
        self._test_acosh(x.astype(np.float32), rtol=1.0e-6)

    @data(*real_inputs)
    def test_acosh_float64(self, x):
        self._test_acosh(x.astype(np.float64), rtol=1.0e-15)

    @data(*safe_complex_inputs)
    def test_acosh_complex64(self, x):
        self._test_acosh(x.astype(np.complex64), rtol=1.0e-5)

    @data(*safe_complex_inputs)
    def test_acosh_complex128(self, x):
        self._test_acosh(x.astype(np.complex128), rtol=1.0e-14)

@ddt
class TestArcsinh(TestMklUFuncs):

    def _test_asinh(self, x, rtol):
        self._test_unary_func(np.arcsinh, mkl.arcsinh, x, rtol)

    @data(*real_inputs)
    def test_asinh_float32(self, x):
        self._test_asinh(x.astype(np.float32), rtol=1.0e-6)

    @data(*real_inputs)
    def test_asinh_float64(self, x):
        self._test_asinh(x.astype(np.float64), rtol=1.0e-15)

    @data(*safe_complex_inputs)
    def test_asinh_complex64(self, x):
        self._test_asinh(x.astype(np.complex64), rtol=5.0e-5) #lnx p27 np19

    @data(*safe_complex_inputs)
    def test_asinh_complex128(self, x):
        self._test_asinh(x.astype(np.complex128), rtol=1.0e-13) #lnx p27 np19

@ddt
class TestArctanh(TestMklUFuncs):

    def _test_atanh(self, x, rtol):
        self._test_unary_func(np.arctanh, mkl.arctanh, x, rtol, sym=True)

    @data(*real_inputs)
    def test_atanh_float32(self, x):
        self._test_atanh(x.astype(np.float32), rtol=1.0e-5) # win64 p35 np19

    @data(*real_inputs)
    def test_atanh_float64(self, x):
        self._test_atanh(x.astype(np.float64), rtol=5.0e-11) #win32 p34 np110

    @data(*safe_complex_inputs)
    def test_atanh_complex64(self, x):
        self._test_atanh(x.astype(np.complex64), rtol=1.0e-4) #centos5x64b

    @data(*safe_complex_inputs)
    def test_atanh_complex128(self, x):
        self._test_atanh(x.astype(np.complex128), rtol=5.0e-14) #win64 p35 np19

test_cases = (TestAdd, TestSubtract, TestSquare, TestMultiply,
    TestAbsolute, TestReciprocal, TestTrueDivide, TestSqrt, TestPower,
    TestHypot, TestExp, TestExpm1, TestLog, TestLog10, TestLog1p,
    TestCos, TestSin, TestTan, TestArccos, TestArcsin, TestArctan2,
    TestCosh, TestSinh, TestTanh, TestArccosh, TestArcsinh, TestArctanh)

if __name__ == '__main__':
    unittest.main()
